package eu.jonahbauer.raytracing.scene.transform;

import eu.jonahbauer.raytracing.math.AABB;
import eu.jonahbauer.raytracing.math.Ray;
import eu.jonahbauer.raytracing.math.Vec3;
import eu.jonahbauer.raytracing.scene.HitResult;
import eu.jonahbauer.raytracing.scene.Hittable;
import eu.jonahbauer.raytracing.scene.Target;
import org.jetbrains.annotations.NotNull;

import java.util.random.RandomGenerator;

public sealed class Translate extends Transform {
    protected final @NotNull Vec3 offset;
    private final @NotNull AABB bbox;

    public static @NotNull Translate create(@NotNull Hittable object, @NotNull Vec3 offset) {
        if (object instanceof Target) {
            return new TranslateTarget(object, offset);
        } else {
            return new Translate(object, offset);
        }
    }

    private Translate(@NotNull Hittable object, @NotNull Vec3 offset) {
        super(object);
        this.offset = offset;

        var bbox = object.getBoundingBox();
        this.bbox = new AABB(
                bbox.min().plus(offset),
                bbox.max().plus(offset)
        );
    }

    @Override
    protected final @NotNull Ray transform(@NotNull Ray ray) {
        return new Ray(ray.origin().minus(offset), ray.direction());
    }

    @Override
    protected final @NotNull HitResult transform(@NotNull HitResult result) {
        return result.withPositionAndNormal(result.position().plus(offset), result.normal());
    }

    @Override
    public final @NotNull AABB getBoundingBox() {
        return bbox;
    }

    @Override
    public @NotNull String toString() {
        return object + " translated by " + offset;
    }

    private static final class TranslateTarget extends Translate implements Target {
        private TranslateTarget(@NotNull Hittable object, @NotNull Vec3 offset) {
            super(object, offset);
        }

        @Override
        public double getProbabilityDensity(@NotNull Vec3 origin, @NotNull Vec3 direction) {
            if (!(object instanceof Target target)) throw new UnsupportedOperationException();
            return target.getProbabilityDensity(origin.minus(offset), direction);
        }

        @Override
        public @NotNull Vec3 getTargetingDirection(@NotNull Vec3 origin, @NotNull RandomGenerator random) {
            if (!(object instanceof Target target)) throw new UnsupportedOperationException();
            return target.getTargetingDirection(origin.minus(offset), random);
        }
    }
}
